Imports System.Collections
Imports System.Net.Http.Formatting
Imports System.Net.Http.Headers
Imports System.Reflection
Imports System.Text
Imports System.Threading.Tasks
Imports System.Xml
Imports System.Xml.Serialization

Namespace HalMediaType
	Public Class HalXmlMediaTypeFormatter
		Inherits BufferedMediaTypeFormatter

		Public Sub New()
			MyBase.New()
			Me.SupportedMediaTypes.Add(New MediaTypeHeaderValue("application/hal"&"xml"))
		End Sub

		Public Overrides Function CanReadType(ByVal type As Type) As Boolean
			Return type.BaseType Is GetType(LinkedResource) OrElse type.BaseType.GetGenericTypeDefinition Is GetType(LinkedResourceCollection(Of ))
		End Function

		Public Overrides Function CanWriteType(ByVal type As Type) As Boolean
			Return type.BaseType Is GetType(LinkedResource) OrElse type.BaseType.GetGenericTypeDefinition Is GetType(LinkedResourceCollection(Of ))
		End Function

		Public Overrides Function ReadFromStream(ByVal type As Type, ByVal readStream As System.IO.Stream, ByVal content As System.Net.Http.HttpContent, ByVal formatterLogger As IFormatterLogger) As Object
			Dim value = CType(Activator.CreateInstance(type), LinkedResource)

			Dim reader = XmlReader.Create(readStream)

			If TypeOf value Is IEnumerable Then
				Dim collection = CType(value, ILinkedResourceCollection)

				reader.ReadStartElement("resource")
				value.HRef = reader.GetAttribute("href")

				Dim innerType = type.BaseType.GetGenericArguments.First

				Do While reader.Read AndAlso reader.LocalName = "resource"
					Dim innerResource = DeserializeInnerResource(reader, innerType)
					collection.Add(innerResource)
				Loop
			Else
				value = DeserializeInnerResource(reader, type)
			End If

			reader.Close()

			Return value
		End Function

		Public Overrides Sub WriteToStream(ByVal type As Type, ByVal value As Object, ByVal writeStream As System.IO.Stream, ByVal content As System.Net.Http.HttpContent)
			Dim settings = New XmlWriterSettings
			settings.Indent = False
			settings.OmitXmlDeclaration = False

			Dim writer = XmlWriter.Create(writeStream, settings)

			Dim resource = CType(value, LinkedResource)

			If TypeOf resource Is IEnumerable Then
				writer.WriteStartElement("resource")
				writer.WriteAttributeString("href", resource.HRef)

				For Each innerResource As LinkedResource In CType(resource, IEnumerable)
					SerializeInnerResource(writer, innerResource)
				Next innerResource

				writer.WriteEndElement()
			Else
				SerializeInnerResource(writer, resource)
			End If

			writer.Flush()
			writer.Close()
		End Sub

		Private Function DeserializeInnerResource(ByVal reader As XmlReader, ByVal innerType As Type) As LinkedResource
			Dim resource = CType(Activator.CreateInstance(innerType), LinkedResource)

			reader.ReadStartElement("resource")
			resource.HRef = reader.GetAttribute("href")

            Dim properties = resource.GetType.GetProperties(BindingFlags.Public Or BindingFlags.Instance).Where(Function(p) p.Name <> "HRef" AndAlso p.Name <> "Links").ToArray

			Do While reader.Read
				If reader.LocalName = "link" Then
					Dim link = New Link
					link.HRef = reader.GetAttribute("href")
					link.Rel = reader.GetAttribute("rel")

                    resource.privateLinks.Add(link)
				ElseIf reader.IsStartElement("resource") Then
					Dim rel = reader.GetAttribute("rel")
					Dim property_ext = properties.FirstOrDefault(Function(p) p.Name Is rel)
					If property_ext IsNot Nothing Then
						Dim propertyValue = DeserializeInnerResource(reader, property_ext.PropertyType)
						property_ext.SetValue(resource, propertyValue)
					End If
				ElseIf reader.IsStartElement Then
					Dim property_ext = properties.FirstOrDefault(Function(p) p.Name = reader.LocalName)
					If property_ext IsNot Nothing Then
						Dim propertyValue = reader.ReadElementContentAs(property_ext.PropertyType, Nothing)
						property_ext.SetValue(resource, propertyValue)
					End If
				End If
			Loop

			Return resource
		End Function

		Private Sub SerializeInnerResource(ByVal writer As XmlWriter, ByVal resource As LinkedResource, Optional ByVal rel As String = Nothing)
			writer.WriteStartElement("resource")
			writer.WriteAttributeString("href", resource.HRef)
			If rel IsNot Nothing Then
				writer.WriteAttributeString("rel", rel)
			End If

            For Each link In resource.privateLinks
                writer.WriteStartElement("link")
                writer.WriteAttributeString("rel", link.Rel)
                writer.WriteAttributeString("href", link.HRef)
                writer.WriteEndElement()
            Next link

            Dim properties = resource.GetType.GetProperties(BindingFlags.Public Or BindingFlags.Instance).Where(Function(p) p.Name <> "HRef" AndAlso p.Name <> "Links")

			For Each property_ext In properties
				Dim propertyValue = property_ext.GetValue(resource)
				If property_ext.PropertyType.IsPrimitive OrElse property_ext.PropertyType Is GetType(String) Then
					writer.WriteElementString(property_ext.Name, propertyValue.ToString)
				ElseIf TypeOf propertyValue Is LinkedResource Then
					SerializeInnerResource(writer, CType(propertyValue, LinkedResource), property_ext.Name)
				End If
			Next property_ext

			writer.WriteEndElement()
		End Sub
	End Class
End Namespace
